/// ____________________________________________________________________ ///
///                                                                      ///
/// BusyFit 0.1 (main.cpp) - Busy Function fitting programme             ///
/// Copyright (C) 2013 Tobias Westmeier                                  ///
/// ____________________________________________________________________ ///
///                                                                      ///
/// Address:  Tobias Westmeier                                           ///
///           ICRAR M468                                                 ///
///           The University of Western Australia                        ///
///           35 Stirling Highway                                        ///
///           Crawley WA 6009                                            ///
///           Australia                                                  ///
///                                                                      ///
/// E-mail:   tobias.westmeier@uwa.edu.au                                ///
/// ____________________________________________________________________ ///
///                                                                      ///
/// This program is free software: you can redistribute it and/or modify ///
/// it under the terms of the GNU General Public License as published by ///
/// the Free Software Foundation, either version 3 of the License, or    ///
/// (at your option) any later version.                                  ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program. If not, see http://www.gnu.org/licenses/.   ///
/// ____________________________________________________________________ ///
///                                                                      ///

#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <cstdlib>
#include <sys/time.h>
// ### WARNING: sys/time.h needed for gettimeofday(), but only works in Linux/Unix!!!

#include "helperFunctions.h"
#include "BusyFit.h"

int main(int argc, char *argv[]);
int printHelpMessage();

int main(int argc, char *argv[])
{
    std::string filename;
    std::vector <double> spectrum;
    std::vector <double> uncertainties;
    double init_par[BUSYFIT_FREE_PARAM] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};           // Initial estimates of parameters.
    bool   mask[BUSYFIT_FREE_PARAM]     = {true, true, true, true, true, true, true};    // Which parameters are free (true) or fixed (false)?
    bool   initialGuess = false;
    bool   noPlot  = false;
    bool   relax   = false;
    bool   verbose = false;
    bool   flagNoise = false;
    int    iterations = 0;                  // Iterations for parameter variation method; 0 = use error propagation method instead (default).
    int    column = 1;                      // Default: second column.
    double noise  = 0.01;                   // Default rms = 0.01.
    int    order  = 2;
    //bool   autoOrder = true;
    
    timeval t1;
    gettimeofday(&t1, NULL);
    srand(t1.tv_usec * t1.tv_sec);          // Initialise random number generator.
    //srand(time(NULL)); Don't use this, only operates on 1 s precision!!!
    
    if(argc < 2)
    {
        std::string message = "\nUsage: busyfit [OPTIONS] FILE\n\n";
        
        message.append("Available options:\n\n");
        message.append("  -c column    Column number of data values in input file. Default is 2.\n\n");
        message.append("  -h, -help    Print a more detailed help message and exit.\n\n");
        message.append("  -n rms       The rms noise level of the spectrum. Default is 0.01.\n\n");
        message.append("  -noplot      Do not plot the fit results on the screen. Useful when \n");
        message.append("               used in batch mode or when gnuplot is unavailable.\n\n");
        message.append("  -o order     Order of the polynomial component (2 or 4). Default is 2.\n\n");
        message.append("  -p P1...Pn   Initial estimates of the free parameters, A, B₁, B₂, C, \n");
        message.append("               XE₀, XP₀, and W, separated by spaces. If not specified, \n");
        message.append("               initial parameters will automatically be determined. \n");
        message.append("               Adding \"f\" at the end will fix a parameter, e.g. \"2.5f\".\n\n");
        message.append("  -relax       Do not check for negative or shifted polynomial component.\n\n");
        message.append("  -u iter      Apply Monte-Carlo method to calculate uncertainties for \n");
        message.append("               observational parameters, using <iter> iterations. Default \n");
        message.append("               is to use a much faster error propagation method.\n\n");
        message.append("  -v           Use verbose mode; show progress information during fitting.\n");
        
        std::cerr << message << std::endl;
        return 0;
    }
    
    for(unsigned int i = 1; i < argc; i++)
    {
        if(std::string(argv[i]) == "-h" or std::string(argv[i]) == "-help")
        {
            // Print help message (using 'less') and exit:
            return printHelpMessage();
        }
        else if(std::string(argv[i]) == "-c")
        {
            if(i < argc - 1)
            {
                i++;
                column = stringToNumber<int>(argv[i]) - 1;
            }
            else
            {
                std::cerr << "Warning (BusyFit): Column number missing; using first column instead." << std::endl;
            }
        }
        else if(std::string(argv[i]) == "-n")
        {
            if(i < argc - 1)
            {
                i++;
                noise = stringToNumber<double>(argv[i]);
                flagNoise = true;
            }
            else
            {
                std::cerr << "Warning (BusyFit): Noise value missing; using default value of 0.01." << std::endl;
            }
        }
        else if(std::string(argv[i]) == "-o")
        {
            if(i < argc - 1)
            {
                i++;
                order = stringToNumber<int>(argv[i]);
            }
            else
            {
                std::cerr << "Warning (BusyFit): No polynomial order specified; using default of 2." << std::endl;
            }
        }
        else if(std::string(argv[i]) == "-p")
        {
            if(i < argc - BUSYFIT_FREE_PARAM)
            {
                for(int j = 0; j < BUSYFIT_FREE_PARAM; j++)
                {
                    std::string parStr = argv[i + j + 1];
                    init_par[j] = stringToNumber<double>(parStr);
                    if(parStr.substr(parStr.size() - 1, 1) == "f") mask[j] = 0;
                }
                
                initialGuess = true;
                
                i += BUSYFIT_FREE_PARAM;
            }
            else std::cerr << "Warning (BusyFit): Initial parameters incomplete; " << BUSYFIT_FREE_PARAM << " expected." << std::endl;
        }
        else if(std::string(argv[i]) == "-noplot")
        {
            noPlot = true;
        }
        else if(std::string(argv[i]) == "-relax")
        {
            relax = true;
        }
        else if(std::string(argv[i]) == "-v")
        {
            verbose = true;
        }
        else if(std::string(argv[i]) == "-u")
        {
            if(i < argc - 1)
            {
                i++;
                iterations = stringToNumber<int>(argv[i]);
                
                if(iterations < 1000)
                {
                    iterations = 1000;
                    std::cerr << "Warning (BusyFit): Too few iterations; adjusting value to 1000." << std::endl;
                }
            }
            else
            {
                iterations = 10000;
                std::cerr << "Warning (BusyFit): Iterations missing; using default value of 10000." << std::endl;
            }
        }
        else
        {
            filename = argv[i];
        }
    }
    
    if(filename.empty())
    {
        std::cerr << "Error (BusyFit): No input file name specified." << std::endl;
        return 1;
    }
    
    std::ifstream fp(filename.c_str(), std::ifstream::in);
    
    if(fp)
    {
        std::string line;
        std::vector <std::string> tokens;
        std::string firstChar;
        
        double addNoise = 0.0;
        
        while(fp.good())
        {
            getline(fp, line);               // Read a line
            stringTrim(line);                // Trim spaces
            stringTok(line, tokens, " \t");  // Tokenise line
            
            if(tokens.size() > column and column >= 0)
            {
                firstChar = (tokens[0]).substr(0, 1);
                
                if(firstChar != "#" and firstChar != "/" and firstChar != "$" and firstChar != "%")
                {
                    // Add some extra noise, if desired (normally disabled):
                    if(0)
                    {
                        double r1 = static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
                        double r2 = static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
                        addNoise = 0.05 * sqrt(-2.0 * log(r1)) * cos(2.0 * MATH_CONST_PI * r2);
                        noise = 0.05;
                    }
                    
                    spectrum.push_back(stringToNumber<double>(tokens[column]) + addNoise);         // addNoise should normally be set to zero! It can be
                    uncertainties.push_back(noise);                                                // activated to add additional noise for testing purposes.
                }
            }
            
            tokens.clear();
        }
        
        fp.close();
    }
    else
    {
        std::cerr << "Error (BusyFit): Failed to open input file \'" << filename << "\'." << std::endl;
        return 1;
    }
    
    
    
    double parValues[BUSYFIT_FREE_PARAM];
    double parUncert[BUSYFIT_FREE_PARAM];
    double chi2  = 0.0;
    double posX  = 0.0;
    double w50   = 0.0;
    double w20   = 0.0;
    double Fint  = 0.0;
    double Fpeak = 0.0;
    double stddev_posX  = 0.0;
    double stddev_w50   = 0.0;
    double stddev_w20   = 0.0;
    double stddev_Fint  = 0.0;
    double stddev_Fpeak = 0.0;
    int    success = 0;
    
    // Run the fitting procedure:
    BusyFit busyFit;
    busyFit.setup(spectrum.size(), &spectrum[0], &uncertainties[0], order, noPlot, relax, verbose);
    busyFit.setFreeParameters(mask);
    
    if(initialGuess == true) success = busyFit.fit(init_par[0], init_par[1], init_par[2], init_par[3], init_par[4], init_par[5], init_par[6]);
    else success = busyFit.fit();
    
    busyFit.getResult(&parValues[0], &parUncert[0], chi2);
    busyFit.getParametersUncertainties(posX, w50, w20, Fpeak, Fint, stddev_posX, stddev_w50, stddev_w20, stddev_Fpeak, stddev_Fint, iterations);
    //else busyFit.getParameters(posX, w50, w20, Fpeak, Fint);
    
    std::cout << std::setprecision(4) << "Pos  =\t" << posX  << "  ±  " << stddev_posX  << "\t [chan]\n";
    std::cout << std::setprecision(4) << "w50  =\t" << w50   << "  ±  " << stddev_w50   << "\t [chan]\n";
    std::cout << std::setprecision(4) << "w20  =\t" << w20   << "  ±  " << stddev_w20   << "\t [chan]\n";
    std::cout << std::setprecision(4) << "Fpk  =\t" << Fpeak << "  ±  " << stddev_Fpeak << "\t [flux]\n";
    std::cout << std::setprecision(4) << "Fint =\t" << Fint  << "  ±  " << stddev_Fint  << "\t [flux × chan]\n" << std::endl;
    
    std::cout << "WARNING: Calculation of parameter uncertainties is experimental.\n" << std::endl;
    if(flagNoise == false) std::cout << "WARNING: No noise level specified (-n); all uncertainties are arbitrary!\n" << std::endl;
    
    // Append results to output file:
    bool historyFileExists = false;
    std::ifstream tmpFile("busyfit_history.txt");
    
    if(tmpFile.is_open())
    {
        historyFileExists = true;
        tmpFile.close();
    }
    
    std::ofstream outputFile;
    outputFile.open("busyfit_history.txt", std::ios::app);
    
    if(outputFile.is_open())
    {
        if(historyFileExists == false)
        {
            outputFile << "# This is the local busyfit history file in which all fit results are stored.\n";
            outputFile << "# In order to delete the history, e.g. prior to a batch run of busyfit on a\n";
            outputFile << "# large number of spectra, simply delete the file; it will be recreated \n";
            outputFile << "# during the next launch of the software. Alternatively, you may choose to \n";
            outputFile << "# remove individual entries from the file. The results of new fits will be \n";
            outputFile << "# appended at the end of the file.\n#\n";
            outputFile << "# The parameters in the respective columns  (separated by tabs) are:\n#\n";
            outputFile << "# Filename\tSuccess\tChi^2\tOrder\tA\tdA\tB_1\tdB_1\tB_2\tdB_2\tC\tdC\tXE_0\tdXE_0\tXP_0\tdXP_0\tW\tdW\tX\tdX\tW_50\tdW_50\tW_20\tdW_20\tF_peak\tdF_peak\tF_int\tdF_int\n#\n";
        }
        
        size_t subStrPos = filename.find_last_of("/");
        
        if(subStrPos < filename.size() - 1) filename = filename.substr(subStrPos + 1);
        
        outputFile << filename << '\t' << success << '\t' << chi2 << '\t' << order;
        
        for(size_t i = 0; i < BUSYFIT_FREE_PARAM; i++)
        {
            outputFile << '\t' << parValues[i] << '\t' << parUncert[i];
        }
        
        outputFile << '\t' << posX << '\t' << stddev_posX << '\t' << w50 << '\t' << stddev_w50 << '\t' << w20 << '\t' << stddev_w20 << '\t' << Fpeak << '\t' << stddev_Fpeak << '\t' << Fint << '\t' << stddev_Fint << '\n';
        
        outputFile.close();
    }
    else
    {
        std::cerr << "Error (BusyFit): Failed to write output file." << std::endl;
        return 1;
    }
    
    return 0;
}

int printHelpMessage()
{
    std::string message;
    
    message.append("\n");
    message.append("Usage: busyfit [OPTIONS] FILE\n");
    message.append("\n");
    message.append("  Note that there is no specific order for the options or file name.\n");
    message.append("\n");
    message.append("\n");
    message.append("Available options:\n");
    message.append("\n");
    message.append("  -c column    Column number of flux values in input file. Default is 2.\n");
    message.append("\n");
    message.append("  -h, -help    Print this detailed help message and exit.\n");
    message.append("\n");
    message.append("  -n rms       The rms noise level of the spectrum. Default is 0.01.\n");
    message.append("               It is important to provide an accurate estimate of the \n");
    message.append("               noise, as otherwise the calculated uncertainties and χ² \n");
    message.append("               values of the fit will be arbitrary and meaningless.\n");
    message.append("\n");
    message.append("  -noplot      Do not plot the fit results on the screen. Useful when \n");
    message.append("               used in batch mode or when gnuplot is unavailable.\n");
    message.append("\n");
    message.append("  -o order     Order of the polynomial component (2 or 4). Default is 2.\n");
    message.append("\n");
    message.append("  -p P1...Pn   Initial estimates of the free parameters, A, B₁, B₂, C, \n");
    message.append("               XE₀, XP₀, and W, separated by spaces. If not specified, \n");
    message.append("               initial parameters will automatically be determined. \n");
    message.append("               Adding \'f\' at the end will fix a parameter, e.g. \'2.5f\'.\n");
    message.append("\n");
    message.append("  -relax       Do not check for negative or shifted polynomial component. \n");
    message.append("               If this option is set, the fit will not be repeated if the \n");
    message.append("               polynomial is found to be either inverted or significantly \n");
    message.append("               shifted relative to the error functions.\n");
    message.append("\n");
    message.append("  -u iter      Calculate uncertainties for observational parameters (flux, \n");
    message.append("               line width, etc.), using a Monte-Carlo method. This will \n");
    message.append("               generate <iter> realisations of the fitted Busy Function \n");
    message.append("               and can therefore be very slow. Default is to use a much \n");
    message.append("               faster error propagation method to compute uncertainties.\n");
    message.append("\n");
    message.append("  -v           Use verbose mode; show progress information during fitting.\n");
    message.append("\n");
    message.append("\n");
    message.append("Introduction:\n");
    message.append("\n");
    message.append("  The busyfit programme  implements a Levenberg–Marquardt χ² minimisation al- \n");
    message.append("  gorithm to fit an asymmetric version of the Busy Function to the integrated \n");
    message.append("  HI spectrum of a galaxy. The normal Busy Function is of the form \n");
    message.append("\n");
    message.append("    B(x) = (A/4) × [erf(B₁ (W + x − XE₀)) + 1] \n");
    message.append("           × [erf(B₂ (W − x + XE₀)) + 1] × [C (x − XP₀)² + 1]\n");
    message.append("\n");
    message.append("  with the free parameters  A, B₁, B₂, C, XE₀, XP₀, and  W.  This form of the \n");
    message.append("  Busy Function is well suited to fitting a wide range of symmetric and asym- \n");
    message.append("  metric double-horn profiles commonly found in galaxies.  In addition to the \n");
    message.append("  fit parameters, the programme will also measure commonly used observational \n");
    message.append("  parameters, including w₅₀ and w₂₀ line widths  as well as peak flux and in- \n");
    message.append("  tegrated flux of the galaxy.\n");
    message.append("\n");
    message.append("  Massive, edge-on spiral galaxies  often produce an integrated spectrum with \n");
    message.append("  very sharp peaks and a fairly broad trough  that is not well described by a \n");
    message.append("  polynomial of order 2. In such cases, busyfit can fit a modified version of \n");
    message.append("  the Busy Function with a polynomial of order 4:\n");
    message.append("\n");
    message.append("    B₂(x) = (A/4) × [erf(B₁ (W + x − XE₀)) + 1] \n");
    message.append("            × [erf(B₂ (W − x + XE₀)) + 1] × [C (x − XP₀)⁴ + 1]\n");
    message.append("\n");
    message.append("  The polynomial order  can be selected  with the \'-o\' option,  that can take \n");
    message.append("  values of 2 or 4. If not specified, a second order is used by default.\n");
    message.append("\n");
    message.append("  If not directed otherwise, the programme will first attempt to fit the full \n");
    message.append("  version of the Busy Function. If the initial fit results in a negative amp- \n");
    message.append("  litude of the polynomial component (C < 0)  or a significant shift with re- \n");
    message.append("  spect to the position of the error functions (|XP₀ − XE₀| > 2 W₀,  where W₀ \n");
    message.append("  is the initial  estimate of the  profile width),  the fit will be discarded \n");
    message.append("  and automatically repeated without the polynomial component, i.e. C and XP₀ \n");
    message.append("  will be fixed to zero  and the fit repeated with only five free parameters. \n");
    message.append("  This behaviour can be changed with the \'-relax\' option, in which case nega- \n");
    message.append("  tive or shifted polynomial solutions will be accepted without repeating the \n");
    message.append("  fit.\n");
    message.append("\n");
    message.append("\n");
    message.append("Input data:\n");
    message.append("\n");
    message.append("  The HI spectrum  will be read  from a text file,  and the user will have to \n");
    message.append("  specify both the input file name  as well as the number  of the column that \n");
    message.append("  contains the flux values of the spectrum. For example, \n");
    message.append("\n");
    message.append("    busyfit -c 3 spectrum.dat\n");
    message.append("\n");
    message.append("  would read the third column from the file spectrum.dat.  The parser assumes \n");
    message.append("  that the tabulated values are separated by either a space or a tab and will \n");
    message.append("  ignore lines starting with a comment character (#, /, $, and %)  as well as \n");
    message.append("  empty lines containing only spaces.\n");
    message.append("\n");
    message.append("  Note that the programme makes the implicit assumption  that the flux values \n");
    message.append("  in the input file  are listed in the correct order and on a regular grid in \n");
    message.append("  either frequency or velocity. The programme will not read the spectral axis \n");
    message.append("  and instead assign the flux values  to spectral channels on a regular grid. \n");
    message.append("  Channel numbers start from zero (i.e. the first encountered flux value will \n");
    message.append("  be assigned  to channel 0),  and it is the user\'s responsibility to convert \n");
    message.append("  the results back into the appropriate spectral units.\n");
    message.append("\n");
    message.append("\n");
    message.append("Initial parameter estimates:\n");
    message.append("\n");
    message.append("  Without any further specifications,  busyfit will automatically determine a \n");
    message.append("  set of initial estimates for the free parameters.  In many cases, this will \n");
    message.append("  be sufficient for the fit to converge. Alternatively, initial estimates can \n");
    message.append("  be provided with the \'-p\' option. For example,\n");
    message.append("\n");
    message.append("    busyfit -p 1 0.5 0.5 0.1 20 20 5 -c 3 spectrum.dat\n");
    message.append("\n");
    message.append("  will set initial parameter estimates of A = 1, B₁ = B₂ = 0.5, C = 0.1, etc. \n");
    message.append("  In addition, parameters can be fixed by adding the letter \'f\' to the value. \n");
    message.append("  For example, \'2.5f\' will set the respective parameter to a value of 2.5 and \n");
    message.append("  exclude the parameter from the fitting procedure.\n");
    message.append("\n");
    message.append("\n");
    message.append("Plotting the results:\n");
    message.append("\n");
    message.append("  The programme will automatically  plot the result of the fit on the screen, \n");
    message.append("  using gnuplot.  This will only work if gnuplot is installed and can be pre- \n");
    message.append("  vented by using the \'-noplot\' option. The plot will show the original spec- \n");
    message.append("  trum, the fitted Busy Function, and the resulting residuals. Note that each \n");
    message.append("  call of busyfit  will open a separate  plot window,  allowing results to be \n");
    message.append("  compared with previous fits. Hence, when automatically fitting a large num- \n");
    message.append("  ber of spectra,  it is important to use the \'-noplot\' option,  as otherwise \n");
    message.append("  a large number of gnuplot windows would be crated, potentially crashing the \n");
    message.append("  window manager.\n");
    message.append("\n");
    message.append("\n");
    message.append("Output files:\n");
    message.append("\n");
    message.append("  Every successful run of busyfit  will create three output files.  The first \n");
    message.append("  two, \'busyfit_output_spectrum.txt\' and  \'busyfit_output_fit.txt\' store  the \n");
    message.append("  original spectrum + residuals and a numerical copy of the fitted Busy Func- \n");
    message.append("  tion, respectively.  These files are used  to plot  the fit results  on the \n");
    message.append("  screen using gnuplot.  The previous contents of the files  will be replaced \n");
    message.append("  whenever the programme is run again.\n");
    message.append("\n");
    message.append("  The third output file, \'busyfit_history.txt\',  will store a copy of the fit \n");
    message.append("  results. Previous results will not be deleted, but instead new results from \n");
    message.append("  the most recent run  of the programme  will be appended  at the end  of the \n");
    message.append("  file.  This behaviour is useful  when running busyfit repeatedly, e.g. on a \n");
    message.append("  large number of spectra.  Before the run, the output file can be deleted or \n");
    message.append("  cleared,  and the results of all runs will then be available for processing \n");
    message.append("  after the last run has finished.\n");
    message.append("\n");
    message.append("  The file \'busyfit_history.txt\' will store each set of results in a separate \n");
    message.append("  line, and each line contains the following entries,  separated by tabs,  in \n");
    message.append("  the given order:\n");
    message.append("\n");
    message.append("   • Input file name\n");
    message.append("   • Success flag (0, 1, or 2)\n");
    message.append("   • Reduced χ²\n");
    message.append("   • Order of the polynomial component (2 or 4)\n");
    message.append("   • Values and uncertainties of the parameters A, B₁, B₂, C, XE₀, XP₀, and W\n");
    message.append("   • Values and uncertainties of centroid, w₅₀, w₂₀, peak flux, and integra-\n");
    message.append("     ted flux\n");
    message.append("\n");
    message.append("  The success flag will tell  whether the fit was fully successful (0),  con- \n");
    message.append("  verged successfully  but with some parameters  negative (1),  or completely \n");
    message.append("  failed to converge (2).  The values of centroid, w₅₀ and w₂₀ will be provi- \n");
    message.append("  ded in channels,  the peak flux will be given in the original flux units of \n");
    message.append("  the input file, and the integrated flux  is measured in  flux units × chan- \n");
    message.append("  nels. Please note that the software starts counting at zero.\n");
    message.append("\n");
    message.append("\n");
    message.append("Uncertainties of observational parameters:\n");
    message.append("\n");
    message.append("  By default,  uncertainties of the observational parameters  will be derived \n");
    message.append("  using an error propagation method. This method uses the standard error pro-\n");
    message.append("  pagation law under the assumption that uncertainties  in the  Busy Function \n");
    message.append("  parameters  linearly propagate  into uncertainties  in the derived observa-\n");
    message.append("  tional parameters.  While this method is very fast, its accuracy can be li-\n");
    message.append("  mited in situations where non-linear effects become relevant.\n");
    message.append("\n");
    message.append("  If accuracy is essential,  an alternative Monte-Carlo method can be invoked \n");
    message.append("  with the \'-u iter\' option. In this case, busyfit will generate <iter> real-\n");
    message.append("  isations  of the fitted  Busy Function  by randomly varying  the function’s \n");
    message.append("  free  parameters  and recalculating  the observational  parameters  in each \n");
    message.append("  iteration.  The uncertainty of each parameter is then taken as the standard \n");
    message.append("  deviation about the mean across all iterations.  A sensible number of iter- \n");
    message.append("  ations is 10,000,  which will determine  uncertainties  with an accuracy of \n");
    message.append("  typically a few percent.  The disadvantage  of the  Monte-Carlo approach is \n");
    message.append("  that it can be  very slow if a large number of spectra  need  to be fitted, \n");
    message.append("  and the default error propagation method may be more suitable in this case.\n");
    message.append("\n");
    message.append("  Warning 1: For the error analysis method to work, it is absolutely essen-\n");
    message.append("             tial that the correct baseline noise level be specified with \n");
    message.append("             the \'-n rms\' option, as otherwise the calculated uncertainties \n");
    message.append("             of all parameters will be entirely arbitrary and meaningless.\n");
    message.append("\n");
    message.append("  Warning 2: The error analysis methods implemented in BusyFit are based on \n");
    message.append("             several assumptions, some of which will break down in certain \n");
    message.append("             situations, leading to unrealistic uncertainty estimates. In \n");
    message.append("             particular, the calculated uncertainties of the derived observa-\n");
    message.append("             tional parameters will almost certainly be wrong if the uncer-\n");
    message.append("             tainty, σ, of any of the non-additive free parameters, P, of the \n");
    message.append("             fit does not fulfil the condition of σ << P.\n");
    message.append("\n");
    message.append("\n");
    message.append("Copyright and licence:\n");
    message.append("\n");
    message.append("  © 2013 Tobias Westmeier\n");
    message.append("\n");
    message.append("  This programme is free software:  you can redistribute it  and/or modify it \n");
    message.append("  under the terms of the GNU General Public License as published  by the Free \n");
    message.append("  Software Foundation,  either version 3 of the License,  or (at your option) \n");
    message.append("  any later version.\n");
    message.append("\n");
    message.append("  This programme is distributed in the hope that it will be useful, but WITH- \n");
    message.append("  OUT ANY WARRANTY;  without even the implied warranty of  MERCHANTABILITY or \n");
    message.append("  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  License  for \n");
    message.append("  more details.\n");
    message.append("\n");
    message.append("  You should have  received  a copy of the  GNU General Public License  along \n");
    message.append("  with this programme. If not, see http://www.gnu.org/licenses/.\n");
    message.append("\n");
    message.append("  Contact: Tobias Westmeier\n");
    message.append("           ICRAR M468\n");
    message.append("           The University of Western Australia\n");
    message.append("           35 Stirling Highway\n");
    message.append("           Crawley WA 6009\n");
    message.append("           Australia\n");
    message.append("\n");
    message.append("  E-mail:  tobias.westmeier@uwa.edu.au\n");
    message.append("\n");
    message.append("\n");
    message.append("End of document. Press \'Q\' to exit.\n");
    message.append("\n");
    
    std::string systemCall = "echo \"";
    systemCall.append(message);
    systemCall.append("\" | less");
    
    return system(systemCall.c_str());
}
